#include "driver.h"
// #define DBG_HARDWAREACTION

// Completeren van IRP's:
// http://msdn.microsoft.com/library/default.asp?url=/library/en-us/kmarch/hh/kmarch/IRPs_6314289f-b703-409e-89c9-66c7193cc045.xml.asp

// Voorbeeld verwerking van IRP's:
// http://msdn.microsoft.com/library/default.asp?url=/library/en-us/kmarch/hh/kmarch/IRPs_32e360d7-a644-44ef-aec0-ffafac00b255.xml.asp

NTSTATUS
  DriverEntry(
    IN PDRIVER_OBJECT  DriverObject,
    IN PUNICODE_STRING  RegistryPath
    )
{
	PDEVICE_OBJECT deviceObject;
	PDEVICE_EXTENSION deviceExtension;
	WCHAR wcKernelModeName[] = KERNELMODE_NAME;
	WCHAR wcUserModeName[] = USERMODE_NAME;
	UNICODE_STRING ucKernelModeName, ucUserModeName;
	NTSTATUS status;

	KdPrint(("LPTDMX.SYS: DriverEntry routine: begin.\n"));

	// DriverEntry's Required Responsibilities 1:
	// Geef de ingangen van de standaardroutines van de driver aan.

	// Routines:
	DriverObject->DriverUnload = Unload;
	// Dispatch routines:
	DriverObject->MajorFunction[IRP_MJ_CREATE]			= DispatchCreate;
	DriverObject->MajorFunction[IRP_MJ_CLOSE]			= DispatchClose;
	DriverObject->MajorFunction[IRP_MJ_WRITE]			= DispatchWrite;
	DriverObject->MajorFunction[IRP_MJ_READ]			= DispatchRead;
	DriverObject->MajorFunction[IRP_MJ_DEVICE_CONTROL]	= DispatchDeviceControl;

	// DriverEntry's Required Responsibilities 2:
	// Creer of initialiseer verschillende driverbrede objecten, types of bronnen die de driver gebruikt.

	// Zet de WCHAR array om naar een UNICODE_STRING
	RtlInitUnicodeString(
	  &ucKernelModeName,	// PUNICODE_STRING DestinationString
	  wcKernelModeName		// PCWSTR SourceString
	);

	// Creer het Device Object voor het LPT-DMX device
	status =						// NTSTATUS 
	  IoCreateDevice(
	    DriverObject,				// IN PDRIVER_OBJECT  DriverObject
	    sizeof(DEVICE_EXTENSION),	// IN ULONG  DeviceExtensionSize
	    &ucKernelModeName,			// IN PUNICODE_STRING  DeviceName  OPTIONAL
	    FILE_DEVICE_PARALLEL_PORT,	// IN DEVICE_TYPE  DeviceType
	    FILE_DEVICE_SECURE_OPEN,	// IN ULONG  DeviceCharacteristics
	    FALSE,						// IN BOOLEAN  Exclusive (moet FALSE zijn)
	    &deviceObject				// OUT PDEVICE_OBJECT  *DeviceObject
	    );
	if (NT_SUCCESS(status))
	{
		deviceObject->Flags |= DO_BUFFERED_IO;
		// Initialiseer de velden in de DEVICE_EXTENSION
		deviceExtension = (PDEVICE_EXTENSION)deviceObject->DeviceExtension;
		deviceExtension->deviceObject = deviceObject;

		// Zet de WCHAR array om naar een UNICODE_STRING
		RtlInitUnicodeString(
		  &ucUserModeName,	// PUNICODE_STRING DestinationString
		  wcUserModeName	// PCWSTR SourceString
		);

		// Koppel een Win32 naam aan de Kernel Mode naam middels een symbolic link
		status = // NTSTATUS
		IoCreateSymbolicLink(
		    &ucUserModeName,	// IN PUNICODE_STRING  SymbolicLinkName
		    &ucKernelModeName	// IN PUNICODE_STRING  DeviceName
		    );
		if (!NT_SUCCESS(status))
		{
			// DriverEntry's Required Responsibilities 3:
			// Maak geheugen vrij dat gealloceerd is maar niet langer gebruikt wordt.
			IoDeleteDevice(
			  deviceObject // IN PDEVICE_OBJECT  DeviceObject
			  );
		}

	}

	// Synchroniseer de interface opnieuw
	for (USHORT i = 0; i <= 255; i++)
		if (!writeChar(BASE, COMMAND_SYNC))
			break;

	KdPrint(("LPTDMX.SYS: DriverEntry routine: einde.\n"));

	// DriverEntry's Required Responsibilities 4:
	// Geef een NTSTATUS code terug die aangeeft of de driver succesvol geladen is en
	// verzoeken kan accepteren en verwerken van de PnP manager om devices te configureren, toevoegen en starten.
	return status;
}

VOID
  Unload(
    IN PDRIVER_OBJECT  DriverObject
    )
{
	WCHAR wcUserModeName[] = USERMODE_NAME;
	UNICODE_STRING ucUserModeName;

	KdPrint(("LPTDMX.SYS: Unload routine: begin.\n"));

	// Zorg ervoor dat het apparaat in compatibiliteitsmodus kan werken
	if (writeChar(BASE, COMMAND_FINISH))
	{
	}

	// Releasing Driver-Allocated Resources 1:
	// Indien mogelijk interrupts uitschakelen op het fysieke device.
	// Vervolgens IoDisconnectInterrupt aanroepen.

	// Releasing Driver-Allocated Resources 2:
	// Ervoor zorgen dat andere driveronderdelen geen gebruik kunnen maken van bronnen
	// die de Unload routine wil verwijderen.
	// Bijv. met behulp van:
	// IoStopTimer, KeRemoveQueueDpc
	// Er moet ervoor gezorgd worden dat eventuele "Work Items" voltooid zijn en
	// dat een eventuele door de driver gecreerde systeemthread zichzelf afsluit.

	// Releasing Driver-Allocated Resources 3:
	// Alle device specifieke bronnen vrijgeven die de driver gealloceerd heeft.
	// Bijv. met behulp van:
	// IoDeleteSymbolicLink, IoDeassignArcName, ExFreePool, MmUnmapIoSpace, MmFreeNonCachedMemory,
	// MmFreeContiguousMemory, FreeCommonBuffer, IoAssignResources, IoReportResourceUsage

	// Hier wordt opnieuw een WCHAR array omgezet naar een UNICODE_STRING,
	// omdat op die manier de UNICODE_STRING niet voortdurend in het geheugen hoeft te blijven staan.
	RtlInitUnicodeString(
	  &ucUserModeName,	// PUNICODE_STRING DestinationString
	  wcUserModeName	// PCWSTR SourceString
	);
	IoDeleteSymbolicLink(
	  &ucUserModeName // IN PUNICODE_STRING  SymbolicLinkName
	  );

	// Releasing Driver-Allocated Resources 4:
	// Systeemobjecten en -bronnen die door de DriverEntry of Reinitialize routine zijn aangemaakt
	// binnen de Device Extensie of in de Controller Extensie.
	// Bijv. met behulp van:
	// IoDisconnectInterrupt, ObDereferenceObject, IoDetachDevice

	// Releasing Driver-Allocated Resources 5:
	// Hardwarebronnen vrijgeven die de DriverEntry of Reinitialize routine zijn geclaimd voor
	// fysieke apparaten, in de \Registry\Machine\Hardware\ResourceMap registersleutel.

	// Releasing Driver-Allocated Resources 6:
	// Namen verwijderen die de DriverEntry of Reinitialize routine in de
	// \Registry..\DeviceMap registersleutel had gezet voor apparaten.

	// Releasing Device and Controller Objects:
	// Device Objecten en Controller Objecten vrijgeven.
	// Als de DriverEntry of Reinitialize routine IoGetConfigurationInformation hadden aangeroepen
	// dan moet deze functie weer worden aangeroepen.
	// Tot slot moet nog enige overige door de driver gealloceerde bronnen worden vrijgegeven.

	IoDeleteDevice(
	  DriverObject->DeviceObject // IN PDEVICE_OBJECT  DeviceObject
	  );

	KdPrint(("LPTDMX.SYS: Unload routine: einde.\n"));
}

// AddDevice routine

// DispatchPnp routine

// DispatchPower routine

/* uit DDK: http://msdn.microsoft.com/library/default.asp?url=/library/en-us/kmarch/hh/kmarch/k113_78355a2a-8b64-4fe3-a652-371048c50f93.xml.asp

IRP_MJ_CREATE
Every kernel-mode driver must handle create requests in a DispatchCreate or DispatchCreateClose routine.
When Sent
Receipt of this request indicates that a user-mode protected subsystem, possibly on behalf of an application, has requested a handle for the file object that represents the target device object, or that a higher-level driver is connecting or attaching its device object to the target device object.

Input Parameters
None

Output Parameters
None

Operation
Many device and intermediate drivers merely set STATUS_SUCCESS in the I/O status block of the IRP and complete the create request. However, what a given driver does on receipt of a create request depends on the driver's design. For example, a driver with pageable-image sections, like the system serial driver, maps in its paged-out code and allocates any resources necessary to handle subsequent I/O requests for the user-mode thread that is attempting to open the device for I/O.

*/

NTSTATUS
  DispatchCreate(
    IN PDEVICE_OBJECT  DeviceObject,
    IN PIRP  Irp
    )
{
	KdPrint(("LPTDMX.SYS: DispatchCreate routine: begin.\n"));

	// How to Complete an IRP in a Dispatch Routine 1:
	// Sets the Status and Information members of the IRP's I/O status block with appropriate values
	Irp->IoStatus.Status		= STATUS_SUCCESS;
	Irp->IoStatus.Information	= 0;

	// How to Complete an IRP in a Dispatch Routine 2:
	// Calls IoCompleteRequest with the IRP and the PriorityBoost IO_NO_INCREMENT.
	IoCompleteRequest(
	  Irp,				// IN PIRP  Irp
	  IO_NO_INCREMENT	// IN CCHAR  PriorityBoost
	  );

	KdPrint(("LPTDMX.SYS: DispatchCreate routine: einde.\n"));

	// How to Complete an IRP in a Dispatch Routine 3:
	// Returns the appropriate STATUS_XXX that it already set in the I/O status block.
	return STATUS_SUCCESS;
}

/* uit DDK: http://msdn.microsoft.com/library/default.asp?url=/library/en-us/kmarch/hh/kmarch/k113_bb864c28-6a78-4158-be96-f03e6e23dc74.xml.asp

IRP_MJ_CLOSE
Every driver must handle close requests in a DispatchClose routine, with the possible exception of a driver whose device cannot be disabled or removed from the machine without bringing down the system. A disk driver whose device holds the system page file is an example of such a driver. Note that the driver of such a device also cannot be unloaded dynamically.
When Sent
Receipt of this request indicates that the last handle of the file object that is associated with the target device object has been closed and released. All outstanding I/O requests have been completed or canceled.

Input Parameters
None

Output Parameters
None

Operation
Many device and intermediate drivers merely set STATUS_SUCCESS in the I/O status block of the IRP and complete the close request. However, what a given driver does on receipt of a close request depends on the driver's design. In general, a driver should undo whatever actions it takes on receipt of the IRP_MJ_CREATE request. Device drivers whose device objects are exclusive, such as a serial driver, also can reset the hardware on receipt of a close request.

The IRP_MJ_CLOSE request is not sent in the context of the process that closed the file object handle. If the driver must release process-specific resources, such as user memory, that the driver previously locked or mapped, it must do so in response to an IRP_MJ_CLEANUP request.

*/

NTSTATUS
  DispatchClose(
    IN PDEVICE_OBJECT  DeviceObject,
    IN PIRP  Irp
    )
{
	KdPrint(("LPTDMX.SYS: DispatchClose routine: begin.\n"));

	// How to Complete an IRP in a Dispatch Routine 1:
	// Sets the Status and Information members of the IRP's I/O status block with appropriate values
	Irp->IoStatus.Status		= STATUS_SUCCESS;
	Irp->IoStatus.Information	= 0;

	// How to Complete an IRP in a Dispatch Routine 2:
	// Calls IoCompleteRequest with the IRP and the PriorityBoost IO_NO_INCREMENT.
	IoCompleteRequest(
	  Irp,					// IN PIRP  Irp
	  IO_NO_INCREMENT		// IN CCHAR  PriorityBoost
	  );

	KdPrint(("LPTDMX.SYS: DispatchClose routine: einde.\n"));

	// How to Complete an IRP in a Dispatch Routine 3:
	// Returns the appropriate STATUS_XXX that it already set in the I/O status block.
	return STATUS_SUCCESS;
}

/* uit DDK: http://msdn.microsoft.com/library/default.asp?url=/library/en-us/kmarch/hh/kmarch/k113_887fcb59-e54b-417c-8c40-5e2a4369c722.xml.asp

IRP_MJ_WRITE
Every device driver that transfers data from the system to its device must handle write requests in a DispatchWrite or DispatchReadWrite routine, as must any higher-level driver layered over such a device driver.
When Sent
Any time following the successful completion of a create request.

Possibly, a user-mode application or Win32 component with a handle for the file object representing the target device object has requested a data transfer to the device. Possibly, a higher-level driver has created and set up the write IRP.

Input Parameters
The driver's I/O stack location in the IRP indicates how many bytes to transfer at Parameters.Write.Length.

Some drivers use the value at Parameters.Write.Key to sort incoming write requests into a driver-determined order in the device queue or in a driver-managed internal queue of IRPs. Certain types of drivers also use the value at Parameters.Write.ByteOffset, which indicates the starting offset for the transfer operation.

Depending on whether the underlying device driver sets up the target device object's Flags with DO_BUFFERED_IO or with DO_DIRECT_IO, data is transferred from one of the following:

The buffer at Irp->AssociatedIrp.SystemBuffer, if the driver uses buffered I/O 
The buffer described by the MDL at Irp->MdlAddress, if the underlying device driver uses direct I/O (DMA or PIO) 

Output Parameters
None

Operation
On receipt of a write request, a higher-level driver sets up the I/O stack location in the IRP for the next-lower driver, or it creates and sets up additional IRPs for one or more lower drivers. It can set up its IoCompletion routine, which is optional for the input IRP but required for driver-created IRPs, by calling IoSetCompletionRoutine. Then, the driver passes the request on to the next-lower driver with IoCallDriver.

On receipt of a write request, a device driver transfers data from system memory to its device. The device driver sets the Information field of the I/O status block to the number of bytes transferred when it completes the IRP.

*/

NTSTATUS
  DispatchWrite(
    IN PDEVICE_OBJECT  DeviceObject,
    IN PIRP  Irp
    )
{
	UCHAR nChannels = 0;

	KdPrint(("LPTDMX.SYS: DispatchWrite routine: begin.\n"));

	// verkrijg een pointer naar de huidig geselecteerde IRP Stack Locatie uit de IRP Stack.
	// uit de IRP Stack Locatie kunnen parameters voor de IRP worden opgehaald.
    PIO_STACK_LOCATION ioStackLocation = // PIO_STACK_LOCATION
	IoGetCurrentIrpStackLocation(
	  Irp // IN PIRP  Irp
	  );

	// maak pointer naar buffer
	PUCHAR bPUCHAR = (PUCHAR)(Irp->AssociatedIrp.SystemBuffer);
	// verzend het commando met behulp van de invoerbuffer
	if (writeChar(BASE, COMMAND_BULK_OUTPUT))
	{
		nChannels = OUTPUT_CHANNELCOUNT;
		if (ioStackLocation->Parameters.Write.Length < nChannels)
			nChannels = (UCHAR)(ioStackLocation->Parameters.Write.Length);
		for (USHORT i = 0; i < nChannels; i++)
			if (writeChar(BASE, (UCHAR)i)) // write channel
				if (writeChar(BASE, bPUCHAR[i])) // write value
				{
				}
		if (!writeChar(BASE, COMMAND_SYNC)) // end transfer
		{
		}
	}

	// How to Complete an IRP in a Dispatch Routine 1:
	// Sets the Status and Information members of the IRP's I/O status block with appropriate values
	Irp->IoStatus.Status		= STATUS_SUCCESS;
	Irp->IoStatus.Information	= nChannels;

	// How to Complete an IRP in a Dispatch Routine 2:
	// Calls IoCompleteRequest with the IRP and the PriorityBoost IO_NO_INCREMENT.
	IoCompleteRequest(
	  Irp,					// IN PIRP  Irp
	  IO_PARALLEL_INCREMENT	// IN CCHAR  PriorityBoost
	  );

	KdPrint(("LPTDMX.SYS: DispatchWrite routine: einde.\n"));

	// How to Complete an IRP in a Dispatch Routine 3:
	// Returns the appropriate STATUS_XXX that it already set in the I/O status block.
	return STATUS_SUCCESS;
}

/* uit DDK: http://msdn.microsoft.com/library/default.asp?url=/library/en-us/kmarch/hh/kmarch/k113_293e4be6-562a-43ab-88c7-75828b5bbf6d.xml.asp

IRP_MJ_READ
Every device driver that transfers data from its device to the system must handle read requests in a DispatchRead or DispatchReadWrite routine, as must any higher-level driver layered over such a device driver.
When Sent
Any time following the successful completion of a create request.

Possibly, a user-mode application or Win32 component with a handle for the file object representing the target device object has requested a data transfer from the device. Possibly, a higher-level driver has created and set up the read IRP.

Input Parameters
The driver's I/O stack location in the IRP indicates how many bytes to transfer at Parameters.Read.Length.

Some drivers use the value at Parameters.Read.Key to sort incoming read requests into a driver-determined order in the device queue or in a driver-managed internal queue of IRPs. Certain types of drivers also use the value at Parameters.Read.ByteOffset, which indicates the starting offset for the transfer operation.

Output Parameters
Depending on whether the underlying device driver sets up the target device object's Flags with DO_BUFFERED_IO or with DO_DIRECT_IO, data is transferred into one of the following:

The buffer at Irp->AssociatedIrp.SystemBuffer if the driver uses buffered I/O 
The buffer described by the MDL at Irp->MdlAddress if the underlying device driver uses direct I/O (DMA or PIO) 

Operation
On receipt of a read request, a higher-level driver sets up the I/O stack location in the IRP for the next-lower driver, or it creates and sets up additional IRPs for one or more lower drivers. It can set up its IoCompletion routine, which is optional for the input IRP but required for driver-created IRPs, by calling IoSetCompletionRoutine. Then, the driver passes the request on to the next-lower driver with IoCallDriver.

On receipt of a read request, a device driver transfers data from its device to system memory. The device driver sets the Information field of the I/O status block to the number of bytes transferred when it completes the IRP.

*/

NTSTATUS
  DispatchRead(
    IN PDEVICE_OBJECT  DeviceObject,
    IN PIRP  Irp
    )
{
	UCHAR nChannels = 0;

	KdPrint(("LPTDMX.SYS: DispatchRead routine: begin.\n"));

	// verkrijg een pointer naar de huidig geselecteerde IRP Stack Locatie uit de IRP Stack.
	// uit de IRP Stack Locatie kunnen parameters voor de IRP worden opgehaald.
    PIO_STACK_LOCATION ioStackLocation = // PIO_STACK_LOCATION
	IoGetCurrentIrpStackLocation(
	  Irp // IN PIRP  Irp
	  );

	// maak pointer naar buffer
	PUCHAR bPUCHAR = (PUCHAR)(Irp->AssociatedIrp.SystemBuffer);
	// verzend het commando en zet het resultaat in de uitvoerbuffer
	if (writeChar(BASE, COMMAND_BULK_INPUT))
	{
		nChannels = INPUT_CHANNELCOUNT;
		if (ioStackLocation->Parameters.Read.Length < nChannels)
			nChannels = (UCHAR)(ioStackLocation->Parameters.Read.Length);
		for (USHORT i = 0; i < nChannels; i++)
		{
			UCHAR nibble_low, nibble_high;
			if (writeReadChar(BASE, (UCHAR)i, nibble_low)) // write channel
				if (readChar(BASE, nibble_high)) // read value
					// vul de uitvoerbuffer met het resultaat van het commando
					bPUCHAR[i] = MAKEBYTE(nibble_low, nibble_high);
		}
		if (!writeChar(BASE, COMMAND_SYNC)) // end transfer
		{
		}
	}

	// How to Complete an IRP in a Dispatch Routine 1:
	// Sets the Status and Information members of the IRP's I/O status block with appropriate values
	Irp->IoStatus.Status		= STATUS_SUCCESS;
	Irp->IoStatus.Information	= nChannels;

	// How to Complete an IRP in a Dispatch Routine 2:
	// Calls IoCompleteRequest with the IRP and the PriorityBoost IO_NO_INCREMENT.
	IoCompleteRequest(
	  Irp,					// IN PIRP  Irp
	  IO_PARALLEL_INCREMENT	// IN CCHAR  PriorityBoost
	  );

	KdPrint(("LPTDMX.SYS: DispatchRead routine: einde.\n"));

	// How to Complete an IRP in a Dispatch Routine 3:
	// Returns the appropriate STATUS_XXX that it already set in the I/O status block.
	return STATUS_SUCCESS;
}

/* uit DDK: http://msdn.microsoft.com/library/default.asp?url=/library/en-us/kmarch/hh/kmarch/k113_c1520a23-a317-4dfd-99bb-58263335fd73.xml.asp

IRP_MJ_DEVICE_CONTROL
Every driver whose device objects belong to a particular device type (see Specifying Device Types) is required to support this request in a DispatchDeviceControl routine, if a set of system-defined I/O control codes (IOCTLs) exists for the type. 
Higher-level drivers usually pass these requests on to an underlying device driver. Each device driver in a driver stack is assumed to support this request, along with a set of device type-specific, public or private IOCTLs. For more information about IOCTLs for specific device types, see device type-specific documentation in the DDK.

When Sent
Any time following the successful completion of a create request.

Input Parameters
The I/O control code is contained at Parameters.DeviceIoControl.IoControlCode in the driver's I/O stack location of the IRP.

Other input parameters depend on the I/O control code's value. For more information, see Buffer Descriptions for I/O Control Codes.

Output Parameters
Output parameters depend on the I/O control code's value. For more information, see Buffer Descriptions for I/O Control Codes.

Operation
A driver receives this I/O control code because user-mode thread has called the Microsoft Win32 DeviceIoControl function, or a higher-level kernel-mode driver has set up the request. Possibly, a user-mode driver has called DeviceIoControl, passing in a driver-defined (also called private) I/O control code, to request device- or driver-specific support from a closely coupled, kernel-mode device driver.

On receipt of a device I/O control request, a higher-level driver usually passes the IRP on to the next-lower driver. However, there are some exceptions to this practice. For example, a class driver that has stored configuration information obtained from the underlying port driver might complete certain IOCTL_XXX requests without passing the IRP down to the corresponding port driver.

On receipt of a device I/O control request, a device driver examines the I/O control code to determine how to satisfy the request. For most public I/O control codes, device drivers transfer a small amount of data to or from the buffer at Irp->AssociatedIrp.SystemBuffer.

*/

NTSTATUS
  DispatchDeviceControl(
    IN PDEVICE_OBJECT  DeviceObject,
    IN PIRP  Irp
    )
{
	NTSTATUS ntStatus;
	UCHAR nibble_low, nibble_high;

	KdPrint(("LPTDMX.SYS: DispatchDeviceControl routine: begin.\n"));

	// Bij METHOD_BUFFERED (zoals hier geldt voor elke IOCTL code):

	// Irp->UserBuffer veld bevat het adres van de User mode buffer.

	// Irp->AssociatedIrp.SystemBuffer veld bevat het adres van de Kernel mode pool buffer,
	// die zodanig groot is dat deze de invoerbuffer van de gebruiker bevat en
	// groot genoeg is om de door de gebruiker verstrekte uitvoerbuffer volledig te kunnen vullen.

	// verkrijg een pointer naar de huidig geselecteerde IRP Stack Locatie uit de IRP Stack.
	// uit de IRP Stack Locatie kunnen parameters voor de IRP worden opgehaald.
    PIO_STACK_LOCATION ioStackLocation = // PIO_STACK_LOCATION
	IoGetCurrentIrpStackLocation(
	  Irp // IN PIRP  Irp
	  );

	switch (ioStackLocation->Parameters.DeviceIoControl.IoControlCode)
	{
/*		case IOCTL_LPTDMX_RESYNC:
			KdPrint(("LPTDMX.SYS: IOCTL_LPTDMX_RESYNC: begin.\n"));

			if
				(
					(ioStackLocation->Parameters.DeviceIoControl.InputBufferLength == 0)
					&&
					(ioStackLocation->Parameters.DeviceIoControl.OutputBufferLength == 0)
				)
			{
				for (USHORT i = 0; i <= 255; i++)
					if (!writeChar(BASE, COMMAND_SYNC))
						break;

				ntStatus = STATUS_SUCCESS;
			}
			else
			{
				// buffer aanwezig terwijl deze niet nodig is
				ntStatus = STATUS_INVALID_PARAMETER;
			}

			Irp->IoStatus.Information = 0;

			KdPrint(("LPTDMX.SYS: IOCTL_LPTDMX_RESYNC: einde.\n"));
			break;
*/
/*		case IOCTL_LPTDMX_FINISH:
			KdPrint(("LPTDMX.SYS: IOCTL_LPTDMX_FINISH: begin.\n"));

			if
				(
					(ioStackLocation->Parameters.DeviceIoControl.InputBufferLength == 0)
					&&
					(ioStackLocation->Parameters.DeviceIoControl.OutputBufferLength == 0)
				)
			{
				if (writeChar(BASE, COMMAND_FINISH))
				{
				}

				ntStatus = STATUS_SUCCESS;
			}
			else
				// buffer aanwezig terwijl deze niet nodig is
				ntStatus = STATUS_INVALID_PARAMETER;

			Irp->IoStatus.Information = 0;

			KdPrint(("LPTDMX.SYS: IOCTL_LPTDMX_FINISH: einde.\n"));
			break;
*/
		case IOCTL_LPTDMX_ECHO:
			KdPrint(("LPTDMX.SYS: IOCTL_LPTDMX_ECHO: begin.\n"));

			// controleer buffergrootten
			if
				(
					(ioStackLocation->Parameters.DeviceIoControl.InputBufferLength >= sizeof(IOCTL_LPTDMX_INPUT_IN))
					&&
					(ioStackLocation->Parameters.DeviceIoControl.OutputBufferLength >= sizeof(IOCTL_LPTDMX_INPUT_OUT))
				)
			{
				// maak pointer naar buffer
				PIOCTL_LPTDMX_ECHO_INOUT bPIOCTL_LPTDMX_ECHO_INOUT = (PIOCTL_LPTDMX_ECHO_INOUT)(Irp->AssociatedIrp.SystemBuffer);
				if (writeChar(BASE, COMMAND_ECHO))
					if (writeReadChar(BASE, bPIOCTL_LPTDMX_ECHO_INOUT->character, nibble_low))
						if (readChar(BASE, nibble_high))
							// vul de uitvoerbuffer met het resultaat van het commando;
							// vanaf nu mag de invoerbuffer niet meer gebruikt worden.
							bPIOCTL_LPTDMX_ECHO_INOUT->character = MAKEBYTE(nibble_low, nibble_high);
				ntStatus = STATUS_SUCCESS;
				Irp->IoStatus.Information = sizeof(IOCTL_LPTDMX_INPUT_OUT);
			}
			else
			{
				// buffer opgegeven door de gebruiker is te klein
				ntStatus = STATUS_INVALID_BUFFER_SIZE;
				Irp->IoStatus.Information = 0;
			}

			KdPrint(("LPTDMX.SYS: IOCTL_LPTDMX_ECHO: einde.\n"));
			break;

		case IOCTL_LPTDMX_INPUT:
			KdPrint(("LPTDMX.SYS: IOCTL_LPTDMX_INPUT: begin.\n"));

			// controleer buffergrootten
			if
				(
					(ioStackLocation->Parameters.DeviceIoControl.InputBufferLength >= sizeof(IOCTL_LPTDMX_INPUT_IN))
					&&
					(ioStackLocation->Parameters.DeviceIoControl.OutputBufferLength >= sizeof(IOCTL_LPTDMX_INPUT_OUT))
				)
			{
				// maak pointers naar buffer
				PIOCTL_LPTDMX_INPUT_IN  bPIOCTL_LPTDMX_INPUT_IN  = (PIOCTL_LPTDMX_INPUT_IN)(Irp->AssociatedIrp.SystemBuffer);
				PIOCTL_LPTDMX_INPUT_OUT bPIOCTL_LPTDMX_INPUT_OUT = (PIOCTL_LPTDMX_INPUT_OUT)(Irp->AssociatedIrp.SystemBuffer);
				// verzend het commando met behulp van de invoerbuffer
				if (writeChar(BASE, COMMAND_INPUT))
					if (writeReadChar(BASE, bPIOCTL_LPTDMX_INPUT_IN->channel, nibble_low)) // write channel
						if (readChar(BASE, nibble_high)) // read value
							// vul de uitvoerbuffer met het resultaat van het commando;
							// vanaf nu mag de invoerbuffer niet meer gebruikt worden.
							bPIOCTL_LPTDMX_INPUT_OUT->value = MAKEBYTE(nibble_low, nibble_high);

				ntStatus = STATUS_SUCCESS;
				Irp->IoStatus.Information = sizeof(IOCTL_LPTDMX_INPUT_OUT);
			}
			else
			{
				// buffer opgegeven door de gebruiker is te klein
				ntStatus = STATUS_INVALID_BUFFER_SIZE;
				Irp->IoStatus.Information = 0;
			}

			KdPrint(("LPTDMX.SYS: IOCTL_LPTDMX_INPUT: einde.\n"));
			break;

		case IOCTL_LPTDMX_OUTPUT:
			KdPrint(("LPTDMX.SYS: IOCTL_LPTDMX_OUTPUT: begin.\n"));

			// controleer buffergrootten
			if (ioStackLocation->Parameters.DeviceIoControl.OutputBufferLength == 0)
				if (ioStackLocation->Parameters.DeviceIoControl.InputBufferLength >= sizeof(IOCTL_LPTDMX_OUTPUT_IN))
				{
					// maak pointer naar buffer
					PIOCTL_LPTDMX_OUTPUT_IN bPIOCTL_LPTDMX_OUTPUT_IN = (PIOCTL_LPTDMX_OUTPUT_IN)(Irp->AssociatedIrp.SystemBuffer);
					// verzend het commando met behulp van de invoerbuffer
					if (writeChar(BASE, COMMAND_OUTPUT))
						if (writeChar(BASE, bPIOCTL_LPTDMX_OUTPUT_IN->channel)) // write channel
							if (writeChar(BASE, bPIOCTL_LPTDMX_OUTPUT_IN->value)) // write value
							{
							}

					ntStatus = STATUS_SUCCESS;
				}
				else
					// buffer opgegeven door de gebruiker is te klein
					ntStatus = STATUS_INVALID_BUFFER_SIZE;
			else
				// buffer aanwezig terwijl deze niet nodig is
				ntStatus = STATUS_INVALID_PARAMETER;

			Irp->IoStatus.Information = 0;

			KdPrint(("LPTDMX.SYS: IOCTL_LPTDMX_OUTPUT: einde.\n"));
			break;

		default:
		{
			KdPrint(("LPTDMX.SYS: invalid IOCTL code: begin.\n"));
			ntStatus = STATUS_INVALID_DEVICE_REQUEST;
			Irp->IoStatus.Information = 0;
			KdPrint(("LPTDMX.SYS: invalid IOCTL code: einde.\n"));
			break;
		}
	}
	
	// How to Complete an IRP in a Dispatch Routine 1:
	// Sets the Status and Information members of the IRP's I/O status block with appropriate values
	Irp->IoStatus.Status		= ntStatus;

	// How to Complete an IRP in a Dispatch Routine 2:
	// Calls IoCompleteRequest with the IRP and the PriorityBoost IO_NO_INCREMENT.
	IoCompleteRequest(
	  Irp,					// IN PIRP  Irp
	  IO_PARALLEL_INCREMENT	// IN CCHAR  PriorityBoost
	  );

	KdPrint(("LPTDMX.SYS: DispatchDeviceControl routine: einde.\n"));

	// How to Complete an IRP in a Dispatch Routine 3:
	// Returns the appropriate STATUS_XXX that it already set in the I/O status block.
	return ntStatus;
}

/*
	Pre:	port = het begin-I/O-adres van een parallelle poort;
	Post:	data is naar het DATA register van de parallelle poort
			waarvan port het beginadres is geschreven;
*/
void writeData(IN PUCHAR port, IN UCHAR data)
{
	port += DATA;
	WRITE_PORT_UCHAR(port, data);
}

/*
	Pre:	port = het begin-I/O-adres van een parallelle poort;
	Post:	getBusy = de genverteerde waarde van het NOT_BUSY bit
			uit het STATUS register van de parallelle poort
			waarvan port het beginadres is;
*/
BOOLEAN getBusy(IN PUCHAR port)
{
	port += STATUS;
	return (READ_PORT_UCHAR(port) & STATUS_NOT_BUSY) == 0;
}

/*
	Pre:	port = het begin-I/O-adres van een parallelle poort;
	Post:	de laagste 4 bits van getNibble zijn gelijk aan de waarden van
			respectievelijk ACK, PE, SEL en ERR uit het STATUS register
			van de parallelle poort waarvan port het beginadres is en
			de hoogste 4 bits van getNibble zijn gelijk aan 0;
*/
UCHAR getNibble(IN PUCHAR port)
{
	ULONG s;
	UCHAR result = 0;
	port += STATUS;
	s = READ_PORT_UCHAR(port);
	if ((s & STATUS_ACK) == STATUS_ACK)
		result |= 8;
	if ((s & STATUS_PE)  == STATUS_PE)
		result |= 4;
	if ((s & STATUS_SEL) == STATUS_SEL)
		result |= 2;
	if ((s & STATUS_ERR) == STATUS_ERR)
		result |= 1;
	return result;
}

/*
	Pre:	port = het begin-I/O-adres van een parallelle poort;
	Post:	(als status = TRUE dan is het STROBE bit in het CONTROL register
			van de parallelle poort waarvan port het beginadres is ingeschakeld) en
			(als status = FALSE dan is het STROBE bit in het CONTROL register
			van de parallelle poort waarvan port het beginadres is uitgeschakeld) en
*/
void setStrobe(IN PUCHAR port, IN BOOLEAN status)
{
	port = port + CONTROL;
	if (status)
		WRITE_PORT_UCHAR(port, READ_PORT_UCHAR(port) | CONTROL_STR);
	else
		WRITE_PORT_UCHAR(port, READ_PORT_UCHAR(port) & !CONTROL_STR);
}


/*
	Pre:	port = het begin-I/O-adres van een parallelle poort;
	Post:	(als hardwareAction = TRUE dan is het commando c is naar de parallelle poort geschreven
			waarvan begin-I/O-adres het beginadres is en
			nibble = de bijbehorende resultaatwaarde) en
			(als hardwareAction = FALSE dan was er een probleem met het benaderen van het LPT-DMX
			apparaat op de parallelle poort waarvan begin-I/O-adres het beginadres is);
*/
BOOLEAN hardwareAction(IN PUCHAR port, IN UCHAR c, OUT UCHAR &nibble)
{
#ifdef DBG_HARDWAREACTION
	KdPrint(("LPTDMX.SYS: hardwareAction: begin.\n"));
#endif
	setStrobe(port, FALSE);

	// wacht totdat het apparaat gereed is voor het ontvangen van data
#ifdef DBG_HARDWAREACTION
	KdPrint(("LPTDMX.SYS: hardwareAction: wachten tot apparaat gereed is voor het ontvangen van data.\n"));
#endif
	while (getBusy(port))
		; // doe niets
#ifdef DBG_HARDWAREACTION
	KdPrint(("LPTDMX.SYS: hardwareAction: apparaat is gereed voor het ontvangen van data.\n"));
#endif

	writeData(port, c);
	setStrobe(port, TRUE);

	// wacht totdat het apparaat het STROBE signaal herkent en het BUSY signaal aanzet
#ifdef DBG_HARDWAREACTION
	KdPrint(("LPTDMX.SYS: hardwareAction: wachten tot STROBE signaal herkend wordt.\n"));
#endif
	while (!getBusy(port))
		; // doe niets
#ifdef DBG_HARDWAREACTION
	KdPrint(("LPTDMX.SYS: hardwareAction: STROBE signaal is herkend.\n"));
#endif

	// bevestig aan het apparaat dat de PC heeft waargenomen dat het apparaat BUSY heeft ingeschakeld
	setStrobe(port, FALSE);

	// wacht totdat het apparaat klaar is met verwerken en het BUSY signaal uitzet
#ifdef DBG_HARDWAREACTION
	KdPrint(("LPTDMX.SYS: hardwareAction: wachten tot het apparaat klaar is met verwerken.\n"));
#endif
	while (getBusy(port))
		; // doe niets
#ifdef DBG_HARDWAREACTION
	KdPrint(("LPTDMX.SYS: hardwareAction: het apparaat is klaar met verwerken.\n"));
#endif

	nibble = getNibble(port);
#ifdef DBG_HARDWAREACTION
	KdPrint(("LPTDMX.SYS: hardwareAction: einde.\n"));
#endif
	return TRUE;
}

/*
	Pre:	port = het begin-I/O-adres van een parallelle poort;
	Post:	(als writeChar = TRUE dan is het commando c is naar de parallelle poort geschreven
			waarvan begin-I/O-adres het beginadres is) en
			(als writeChar = FALSE dan was er een probleem met het benaderen van het LPT-DMX
			apparaat op de parallelle poort waarvan begin-I/O-adres het beginadres is);
*/
BOOLEAN writeChar(IN PUCHAR port, IN UCHAR c)
{
	UCHAR dummy;
	BOOLEAN result = hardwareAction(port, c, dummy);
	KdPrint(("LPTDMX.SYS: writeChar: written %d.\n", c));
	return result;
}

/*
	Pre:	port = het begin-I/O-adres van een parallelle poort;
	Post:	(als readChar = TRUE dan is COMMAND_SYNC naar de parallelle poort geschreven
			waarvan begin-I/O-adres het beginadres is en
			nibble = de bijbehorende resultaatwaarde) en
			(als readChar = FALSE dan was er een probleem met het benaderen van het LPT-DMX
			apparaat op de parallelle poort waarvan begin-I/O-adres het beginadres is);
*/
BOOLEAN readChar(IN PUCHAR port, OUT UCHAR &nibble)
{
	BOOLEAN result = hardwareAction(port, COMMAND_SYNC, nibble);
	KdPrint(("LPTDMX.SYS: readChar: read %d.\n", nibble));
	return result;
}

/*
	Pre:	port = het begin-I/O-adres van een parallelle poort;
	Post:	(als writeReadChar = TRUE dan is het commando c is naar de parallelle poort geschreven
			waarvan begin-I/O-adres het beginadres is en
			nibble = de bijbehorende resultaatwaarde) en
			(als writeReadChar = FALSE dan was er een probleem met het benaderen van het LPT-DMX
			apparaat op de parallelle poort waarvan begin-I/O-adres het beginadres is);
*/
BOOLEAN writeReadChar(IN PUCHAR port, IN UCHAR c, OUT UCHAR &nibble)
{
	BOOLEAN result = hardwareAction(port, c, nibble);
	KdPrint(("LPTDMX.SYS: writeReadChar: written %d read %d.\n", c, nibble));
	return result;
}
