extern "C"
{
    #include <ntddk.h>
}
#include "ioctl.h"

#define KERNELMODE_NAME L"\\Device\\LptDmx0"
#define USERMODE_NAME L"\\??\\LPTDMX1"

typedef struct
{
	PDEVICE_OBJECT	deviceObject;
} DEVICE_EXTENSION, *PDEVICE_EXTENSION;

#ifdef DBG
	UCHAR MAKEBYTE(UCHAR nibbleLow, UCHAR nibbleHigh)
	{
		UCHAR byte = (nibbleHigh << 4) | nibbleLow;
		KdPrint(("LPTDMX.SYS: MAKEBYTE: %d", byte));
		return byte;
	}
#endif
#ifndef DBG
	#define MAKEBYTE(nibbleLow, nibbleHigh) \
		(nibbleHigh << 4) | nibbleLow
#endif

//#define ReadData(port) READ_PORT_

#define BASE	(PUCHAR)0x378
#define DATA	0
#define STATUS	1
#define CONTROL	2

#define DATA_D7	128
#define DATA_D6	64
#define DATA_D5	32
#define DATA_D4	16
#define DATA_D3	8
#define DATA_D2	4
#define DATA_D1	2
#define DATA_D0	1

#define STATUS_NOT_BUSY	128 // laag indien actief
#define STATUS_ACK		64
#define STATUS_PE		32
#define STATUS_SEL		16
#define STATUS_ERR		8

#define CONTROL_SELIN	8
#define CONTROL_RESET	4
#define CONTROL_LF		2
#define CONTROL_STR		1 // hoog indien actief

// Basic commands

#define COMMAND_SYNC					255 // gemplementeerd (DriverEntry)
#define COMMAND_FINISH					'^' // gemplementeerd (DriverUnload)
#define COMMAND_RESET					'-'
#define COMMAND_DEFAULTS				'P'
#define COMMAND_CLEAR_ALL				'C'
#define COMMAND_CLEAR_INPUT				'8'
#define COMMAND_CLEAR_OUTPUT			'9'
#define COMMAND_ECHO					'%' // gemplementeerd (DeviceIoControl)
  
// I/O commands

#define COMMAND_INPUT					'i' // gemplementeerd (DeviceIoControl)
#define COMMAND_BULK_INPUT				'I' // gemplementeerd (ReadFile)
#define COMMAND_OUTPUT					'o' // gemplementeerd (DeviceIoControl)
#define COMMAND_BULK_OUTPUT				'O' // gemplementeerd (WriteFile)
#define COMMAND_SET_MASS_OUTPUT_LIMITS	'1'
#define COMMAND_MASS_OUTPUT				'Q'

// Input commands

#define COMMAND_ENABLE_INPUT			'R'
#define COMMAND_DISABLE_INPUT			'r'
#define COMMAND_SET_INPUT_STARTCODE		'S'
#define COMMAND_ENABLE_INPUT_WILDCARD	'*'
#define COMMAND_DISABLE_INPUT_WILDCARD	'#'
#define COMMAND_GET_INPUT_STARTCODE		'K'
#define COMMAND_GET_INPUT_COUNT			'G'
#define COMMAND_GET_TIME_SINCE_INPUT	'Z'

// Output commands

#define COMMAND_ENABLE_OUTPUT			'T'
#define COMMAND_DISABLE_OUTPUT			't'
#define COMMAND_SET_BREAK_TIME			'B' // niet relevant
#define COMMAND_SET_MAB_TIME			'M' // niet relevant
#define COMMAND_SET_OUTPUT_STARTCODE	'D' // niet relevant
#define COMMAND_SET_OUTPUT_COUNT		'L'

// Multiplexer commands

#define COMMAND_ENABLE_MUX				'E'
#define COMMAND_DISABLE_MUX				'e'
#define COMMAND_INVERTED_MUX			'N'
#define COMMAND_NON_INVERTED_MUX		'n'
#define COMMAND_SET_MUX_ADDRESS			'A'
#define COMMAND_FORCE_MUX_OFF			'F'
#define COMMAND_NO_FORCE_MUX_OFF		'f'
#define COMMAND_GET_MUX_VALUE			'Y'

// Merge commands

#define COMMAND_ENABLE_MERGE			'X'
#define COMMAND_DISABLE_MERGE			'x'

// EEPROM commands
  
#define COMMAND_READ_EEPROM				'V'
#define COMMAND_WRITE_EEPROM			'W'

#define INPUT_CHANNELCOUNT  245 // excluding start code
#define OUTPUT_CHANNELCOUNT 255 // excluding start code

VOID
  Unload(
    IN PDRIVER_OBJECT  DriverObject
    );

NTSTATUS
  DispatchCreate(
    IN PDEVICE_OBJECT  DeviceObject,
    IN PIRP  Irp
    );

NTSTATUS
  DispatchClose(
    IN PDEVICE_OBJECT  DeviceObject,
    IN PIRP  Irp
    );

NTSTATUS
  DispatchWrite(
    IN PDEVICE_OBJECT  DeviceObject,
    IN PIRP  Irp
    );

NTSTATUS
  DispatchRead(
    IN PDEVICE_OBJECT  DeviceObject,
    IN PIRP  Irp
    );

NTSTATUS
  DispatchDeviceControl(
    IN PDEVICE_OBJECT  DeviceObject,
    IN PIRP  Irp
    );

BOOLEAN writeReadChar(IN PUCHAR port, IN UCHAR c, OUT UCHAR &nibble);
BOOLEAN writeChar(IN PUCHAR port, IN UCHAR c);
BOOLEAN readChar(IN PUCHAR port, OUT UCHAR &nibble);
